"""Tornado handlers for kernel specifications.

Preliminary documentation at https://github.com/ipython/ipython/wiki/IPEP-25%3A-Registry-of-installed-kernels#rest-api
"""

# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.

import glob
import json
import os
pjoin = os.path.join

from tornado import web

from ...base.handlers import APIHandler
from ...utils import ensure_async, url_path_join, url_unescape



def kernelspec_model(handler, name, spec_dict, resource_dir):
    """Load a KernelSpec by name and return the REST API model"""
    d = {
        'name': name,
        'spec': spec_dict,
        'resources': {}
    }

    # Add resource files if they exist
    resource_dir = resource_dir
    for resource in ['kernel.js', 'kernel.css']:
        if os.path.exists(pjoin(resource_dir, resource)):
            d['resources'][resource] = url_path_join(
                handler.base_url,
                'kernelspecs',
                name,
                resource
            )
    for logo_file in glob.glob(pjoin(resource_dir, 'logo-*')):
        fname = os.path.basename(logo_file)
        no_ext, _ = os.path.splitext(fname)
        d['resources'][no_ext] = url_path_join(
            handler.base_url,
            'kernelspecs',
            name,
            fname
        )
    return d


def is_kernelspec_model(spec_dict):
    """Returns True if spec_dict is already in proper form.  This will occur when using a gateway."""
    return isinstance(spec_dict, dict) and 'name' in spec_dict and 'spec' in spec_dict and 'resources' in spec_dict


class MainKernelSpecHandler(APIHandler):

    @web.authenticated
    async def get(self):
        ksm = self.kernel_spec_manager
        km = self.kernel_manager
        model = {}
        model['default'] = km.default_kernel_name
        model['kernelspecs'] = specs = {}
        kspecs = await ensure_async(ksm.get_all_specs())
        for kernel_name, kernel_info in kspecs.items():
            try:
                if is_kernelspec_model(kernel_info):
                    d = kernel_info
                else:
                    d = kernelspec_model(self, kernel_name, kernel_info['spec'], kernel_info['resource_dir'])
            except Exception:
                self.log.error("Failed to load kernel spec: '%s'", kernel_name, exc_info=True)
                continue
            specs[kernel_name] = d
        self.set_header("Content-Type", 'application/json')
        self.finish(json.dumps(model))


class KernelSpecHandler(APIHandler):

    @web.authenticated
    async def get(self, kernel_name):
        ksm = self.kernel_spec_manager
        kernel_name = url_unescape(kernel_name)
        try:
            spec = await ensure_async(ksm.get_kernel_spec(kernel_name))
        except KeyError as e:
            raise web.HTTPError(404, u'Kernel spec %s not found' % kernel_name) from e
        if is_kernelspec_model(spec):
            model = spec
        else:
            model = kernelspec_model(self, kernel_name, spec.to_dict(), spec.resource_dir)
        self.set_header("Content-Type", 'application/json')
        self.finish(json.dumps(model))


# URL to handler mappings

kernel_name_regex = r"(?P<kernel_name>[\w\.\-%]+)"

default_handlers = [
    (r"/api/kernelspecs", MainKernelSpecHandler),
    (r"/api/kernelspecs/%s" % kernel_name_regex, KernelSpecHandler),
]



import pytest

from tornado import web

from jupyter_server.services.sessions.sessionmanager import SessionManager
from jupyter_server.services.kernels.kernelmanager import MappingKernelManager
from jupyter_server.services.contents.manager import ContentsManager
from jupyter_server._tz import utcnow, isoformat


class DummyKernel(object):
    def __init__(self, kernel_name='python'):
        self.kernel_name = kernel_name


dummy_date = utcnow()
dummy_date_s = isoformat(dummy_date)


class DummyMKM(MappingKernelManager):
    """MappingKernelManager interface that doesn't start kernels, for testing"""
    def __init__(self, *args, **kwargs):
        super(DummyMKM, self).__init__(*args, **kwargs)
        self.id_letters = iter(u'ABCDEFGHIJK')

    def _new_id(self):
        return next(self.id_letters)

    async def start_kernel(self, kernel_id=None, path=None, kernel_name='python', **kwargs):
        kernel_id = kernel_id or self._new_id()
        k = self._kernels[kernel_id] = DummyKernel(kernel_name=kernel_name)
        self._kernel_connections[kernel_id] = 0
        k.last_activity = dummy_date
        k.execution_state = 'idle'
        return kernel_id

    async def shutdown_kernel(self, kernel_id, now=False):
        del self._kernels[kernel_id]


@pytest.fixture
def session_manager():
    return SessionManager(
        kernel_manager=DummyMKM(),
        contents_manager=ContentsManager())


async def create_multiple_sessions(session_manager, *kwargs_list):
    sessions = []
    for kwargs in kwargs_list:
        kwargs.setdefault('type', 'notebook')
        session = await session_manager.create_session(**kwargs)
        sessions.append(session)
    return sessions


async def test_get_session(session_manager):
    session = await session_manager.create_session(
        path='/path/to/test.ipynb', 
        kernel_name='bar',
        type='notebook'
    )
    session_id = session['id']
    model = await session_manager.get_session(session_id=session_id)
    expected = {
        'id':session_id,
        'path': u'/path/to/test.ipynb',
        'notebook': {'path': u'/path/to/test.ipynb', 'name': None},
        'type': 'notebook',
        'name': None,
        'kernel': {
            'id': 'A',
            'name': 'bar',
            'connections': 0,
            'last_activity': dummy_date_s,
            'execution_state': 'idle',
        }
    }
    assert model == expected


async def test_bad_get_session(session_manager):
    session = await session_manager.create_session(
        path='/path/to/test.ipynb',
        kernel_name='foo',
        type='notebook'
    )
    with pytest.raises(TypeError):
        await session_manager.get_session(bad_id=session['id'])


async def test_get_session_dead_kernel(session_manager):
    session = await session_manager.create_session(
        path='/path/to/1/test1.ipynb',
        kernel_name='python',
        type='notebook'
    )
    # Kill the kernel
    await session_manager.kernel_manager.shutdown_kernel(session['kernel']['id'])
    with pytest.raises(KeyError):
        await session_manager.get_session(session_id=session['id'])
    # no session left
    listed = await session_manager.list_sessions()
    assert listed == []


async def test_list_session(session_manager):
    sessions = await create_multiple_sessions(
        session_manager,
        dict(path='/path/to/1/test1.ipynb', kernel_name='python'),
        dict(path='/path/to/2/test2.py', type='file', kernel_name='python'),
        dict(path='/path/to/3', name='foo', type='console', kernel_name='python'),
    )
    sessions = await session_manager.list_sessions()
    expected = [
        {
            'id':sessions[0]['id'],
            'path': u'/path/to/1/test1.ipynb',
            'type': 'notebook',
            'notebook': {'path': u'/path/to/1/test1.ipynb', 'name': None},
            'name': None,
            'kernel': {
                'id': 'A',
                'name':'python',
                'connections': 0,
                'last_activity': dummy_date_s,
                'execution_state': 'idle',
            }
        }, {
            'id':sessions[1]['id'],
            'path': u'/path/to/2/test2.py',
            'type': 'file',
            'name': None,
            'kernel': {
                'id': 'B',
                'name':'python',
                'connections': 0,
                'last_activity': dummy_date_s,
                'execution_state': 'idle',
            }
        }, {
            'id':sessions[2]['id'],
            'path': u'/path/to/3',
            'type': 'console',
            'name': 'foo',
            'kernel': {
                'id': 'C',
                'name':'python',
                'connections': 0,
                'last_activity': dummy_date_s,
                'execution_state': 'idle',
            }
        }
    ]
    assert sessions == expected


async def test_list_sessions_dead_kernel(session_manager):
    sessions = await create_multiple_sessions(
        session_manager,
        dict(path='/path/to/1/test1.ipynb', kernel_name='python'),
        dict(path='/path/to/2/test2.ipynb', kernel_name='python'),
    )
    # kill one of the kernels
    await session_manager.kernel_manager.shutdown_kernel(sessions[0]['kernel']['id'])
    listed = await session_manager.list_sessions()
    expected = [
        {
            'id': sessions[1]['id'],
            'path': u'/path/to/2/test2.ipynb',
            'type': 'notebook',
            'name': None,
            'notebook': {'path': u'/path/to/2/test2.ipynb', 'name': None},
            'kernel': {
                'id': 'B',
                'name':'python',
                'connections': 0,
                'last_activity': dummy_date_s,
                'execution_state': 'idle',
            }
        }
    ]
    assert listed == expected


async def test_update_session(session_manager):
    session = await session_manager.create_session(
        path='/path/to/test.ipynb',
        kernel_name='julia',
        type='notebook'
    )
    session_id = session['id']
    await session_manager.update_session(session_id, path='/path/to/new_name.ipynb')
    model = await session_manager.get_session(session_id=session_id)
    expected = {'id':session_id,
                'path': u'/path/to/new_name.ipynb',
                'type': 'notebook',
                'name': None,
                'notebook': {'path': u'/path/to/new_name.ipynb', 'name': None},
                'kernel': {
                    'id': 'A',
                    'name':'julia',
                    'connections': 0,
                    'last_activity': dummy_date_s,
                    'execution_state': 'idle',
                }
    }
    assert model == expected


async def test_bad_update_session(session_manager):
    # try to update a session with a bad keyword ~ raise error
    session = await session_manager.create_session(
        path='/path/to/test.ipynb',
        kernel_name='ir',
        type='notegbook'
    )
    session_id = session['id']
    with pytest.raises(TypeError):
        await session_manager.update_session(session_id=session_id, bad_kw='test.ipynb') # Bad keyword


async def test_delete_session(session_manager):
    sessions = await create_multiple_sessions(
        session_manager,
        dict(path='/path/to/1/test1.ipynb', kernel_name='python'),
        dict(path='/path/to/2/test2.ipynb', kernel_name='python'),
        dict(path='/path/to/3', name='foo', type='console', kernel_name='python'),
    )
    await session_manager.delete_session(sessions[1]['id'])
    new_sessions = await session_manager.list_sessions()
    expected = [{
            'id': sessions[0]['id'],
            'path': u'/path/to/1/test1.ipynb',
            'type': 'notebook',
            'name': None,
            'notebook': {'path': u'/path/to/1/test1.ipynb', 'name': None},
            'kernel': {
                'id': 'A',
                'name':'python',
                'connections': 0,
                'last_activity': dummy_date_s,
                'execution_state': 'idle',
            }
        }, {
            'id': sessions[2]['id'],
            'type': 'console',
            'path': u'/path/to/3',
            'name': 'foo',
            'kernel': {
                'id': 'C',
                'name':'python',
                'connections': 0,
                'last_activity': dummy_date_s,
                'execution_state': 'idle',
            }
        }
    ]
    assert new_sessions == expected


async def test_bad_delete_session(session_manager):
    # try to delete a session that doesn't exist ~ raise error
    await session_manager.create_session(
        path='/path/to/test.ipynb', 
        kernel_name='python',
        type='notebook'
    )
    with pytest.raises(TypeError):
        await session_manager.delete_session(bad_kwarg='23424') # Bad keyword
    with pytest.raises(web.HTTPError):
        await session_manager.delete_session(session_id='23424') # nonexistent




import os
import sys
import ctypes
import pytest

from traitlets.tests.utils import check_help_all_output
from jupyter_server.utils import url_escape, url_unescape, is_hidden, is_file_hidden, secure_write
from ipython_genutils.py3compat import cast_unicode


def test_help_output():
    check_help_all_output('jupyter_server')



@pytest.mark.parametrize(
    'unescaped,escaped',
    [
        (
            '/this is a test/for spaces/', 
            '/this%20is%20a%20test/for%20spaces/'
        ),
        (
            'notebook with space.ipynb',
            'notebook%20with%20space.ipynb'
        ),
        (
            '/path with a/notebook and space.ipynb',
            '/path%20with%20a/notebook%20and%20space.ipynb'
        ),
        (
            '/ !@$#%^&* / test %^ notebook @#$ name.ipynb',
            '/%20%21%40%24%23%25%5E%26%2A%20/%20test%20%25%5E%20notebook%20%40%23%24%20name.ipynb'
        )
    ]
)
def test_url_escaping(unescaped, escaped):
    # Test escaping.
    path = url_escape(unescaped)
    assert path == escaped
    # Test unescaping.
    path = url_unescape(escaped)
    assert path == unescaped


def test_is_hidden(tmp_path):
    root = str(tmp_path)
    subdir1_path = tmp_path / 'subdir'
    subdir1_path.mkdir()
    subdir1 = str(subdir1_path)
    assert not is_hidden(subdir1, root)
    assert not is_file_hidden(subdir1)

    subdir2_path = tmp_path / '.subdir2'
    subdir2_path.mkdir()
    subdir2 = str(subdir2_path)
    assert is_hidden(subdir2, root)
    assert is_file_hidden(subdir2)

    subdir34_path = tmp_path / 'subdir3' / '.subdir4'
    subdir34_path.mkdir(parents=True)
    subdir34 = str(subdir34_path)
    assert is_hidden(subdir34, root)
    assert is_hidden(subdir34)

    subdir56_path = tmp_path / '.subdir5' / 'subdir6'
    subdir56_path.mkdir(parents=True)
    subdir56 = str(subdir56_path)
    assert is_hidden(subdir56, root)
    assert is_hidden(subdir56)
    assert not is_file_hidden(subdir56)
    assert not is_file_hidden(subdir56, os.stat(subdir56))


@pytest.mark.skipif(sys.platform != "win32", reason="Test is not windows.")
def test_is_hidden_win32(tmp_path):
    root = str(tmp_path)
    root = cast_unicode(root)
    subdir1 = tmp_path / 'subdir'
    subdir1.mkdir()
    assert not is_hidden(str(subdir1), root)
    ctypes.windll.kernel32.SetFileAttributesW(str(subdir1), 0x02)
    assert is_hidden(str(subdir1), root)
    assert is_file_hidden(str(subdir1))




from jupyter_server.services.config.manager import ConfigManager


DEFAULT_SECTION_NAME = "jupyter_server_config"


class ExtensionConfigManager(ConfigManager):
    """A manager class to interface with Jupyter Server Extension config
    found in a `config.d` folder. It is assumed that all configuration
    files in this directory are JSON files.
    """
    def get_jpserver_extensions(
        self,
        section_name=DEFAULT_SECTION_NAME
    ):
        """Return the jpserver_extensions field from all
        config files found."""
        data = self.get(section_name)
        return (
            data
            .get("ServerApp", {})
            .get("jpserver_extensions", {})
        )

    def enabled(
        self,
        name,
        section_name=DEFAULT_SECTION_NAME,
        include_root=True
    ):
        """Is the extension enabled?"""
        extensions = self.get_jpserver_extensions(section_name)
        try:
            return extensions[name]
        except KeyError:
            return False

    def enable(self, name):
        data = {
            "ServerApp": {
                "jpserver_extensions": {
                    name: True
                }
            }
        }
        self.update(name, data)

    def disable(self, name):
        data = {
            "ServerApp": {
                "jpserver_extensions": {
                    name: False
                }
            }
        }
        self.update(name, data)



import pytest
from collections import OrderedDict
from traitlets.tests.utils import check_help_all_output

from jupyter_server.extension.serverextension import (
    toggle_server_extension_python,
    _get_config_dir
)
from jupyter_server.config_manager import BaseJSONConfigManager


# Use ServerApps environment because it monkeypatches
# jupyter_core.paths and provides a config directory
# that's not cross contaminating the user config directory.
pytestmark = pytest.mark.usefixtures("jp_environ")


def test_help_output():
    check_help_all_output('jupyter_server.extension.serverextension')
    check_help_all_output('jupyter_server.extension.serverextension', ['enable'])
    check_help_all_output('jupyter_server.extension.serverextension', ['disable'])
    check_help_all_output('jupyter_server.extension.serverextension', ['install'])
    check_help_all_output('jupyter_server.extension.serverextension', ['uninstall'])


def get_config(sys_prefix=True):
    cm = BaseJSONConfigManager(config_dir=_get_config_dir(sys_prefix=sys_prefix))
    data = cm.get("jupyter_server_config")
    return data.get("ServerApp", {}).get("jpserver_extensions", {})


def test_enable(jp_env_config_path, jp_extension_environ):
    toggle_server_extension_python('mock1', True)
    config = get_config()
    assert config['mock1']


def test_disable(jp_env_config_path, jp_extension_environ):
    toggle_server_extension_python('mock1', True)
    toggle_server_extension_python('mock1', False)

    config = get_config()
    assert not config['mock1']


def test_merge_config(
        jp_env_config_path,
        jp_configurable_serverapp,
        jp_extension_environ
):
    # Toggle each extension module with a JSON config file
    # at the sys-prefix config dir.
    toggle_server_extension_python(
        'tests.extension.mockextensions.mockext_sys',
        enabled=True,
        sys_prefix=True
    )
    toggle_server_extension_python(
        'tests.extension.mockextensions.mockext_user',
        enabled=True,
        user=True
    )

    # Write this configuration in two places, sys-prefix and user.
    # sys-prefix supercedes users, so the extension should be disabled
    # when these two configs merge.
    toggle_server_extension_python(
        'tests.extension.mockextensions.mockext_both',
        enabled=True,
        sys_prefix=True
    )
    toggle_server_extension_python(
        'tests.extension.mockextensions.mockext_both',
        enabled=False,
        user=True
    )

    arg = "--ServerApp.jpserver_extensions={{'{mockext_py}': True}}".format(
        mockext_py='tests.extension.mockextensions.mockext_py'
    )

    # Enable the last extension, mockext_py, using the CLI interface.
    app = jp_configurable_serverapp(
        config_dir=str(jp_env_config_path),
        argv=[arg]
    )
    # Verify that extensions are enabled and merged in proper order.
    extensions = app.jpserver_extensions
    assert extensions['tests.extension.mockextensions.mockext_user']
    assert extensions['tests.extension.mockextensions.mockext_sys']
    assert extensions['tests.extension.mockextensions.mockext_py']
    # Merging should causes this extension to be disabled.
    assert not extensions['tests.extension.mockextensions.mockext_both']


@pytest.mark.parametrize(
    'jp_server_config',
    [
        {
            "ServerApp": {
                "jpserver_extensions": OrderedDict([
                    ('tests.extension.mockextensions.mock2', True),
                    ('tests.extension.mockextensions.mock1', True)
                ])
            }
        }
    ]
)
def test_load_ordered(jp_serverapp, jp_server_config):
    assert jp_serverapp.mockII is True, "Mock II should have been loaded"
    assert jp_serverapp.mockI is True, "Mock I should have been loaded"
    assert jp_serverapp.mock_shared == 'II', "Mock II should be loaded after Mock I"
